import numpy as np
from precipitation import read, sum_by_month
from precipitation import create_bargraph

months = np.arange(12) + 1
d12, d13, d14 = read("p12.dat"), read("p13.dat"), read("p14.dat")
prec12 = sum_by_month(d12, months)
prec13 = sum_by_month(d13, months)
prec14 = sum_by_month(d14, months)

create_bargraph("out.png", months,
	            ["2012", "2013", "2014"], prec12, prec13, prec14)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""IPython Module"""
from __future__ import (absolute_import, print_function,
                        division)


from ..persistence.models import *                                               # pylint: disable=wildcard-import
from ..persistence import persistence_config, relational, content


def init(path=None, ipython=None):
    """Initiate noWorkflow extension.
    Load D3, IPython magics, and connect to database


    Keyword Arguments:
    path -- database path (default=current directory)
    ipython -- IPython object (default=None)
    """

    import os
    from .magics import register_magics
    try:
        from .hierarchymagic import load_ipython_extension as load_hierarchy
        load_hierarchy(ipython)
    except ImportError:
        print("Warning: Sphinx is not installed. Dot "
              "graphs won't work")

    register_magics(ipython)

    if path is None:
        path = os.getcwd()
    persistence_config.connect(path)

    return u"ok"



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Trial Graph Module"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import weakref

from collections import defaultdict

from future.utils import viewitems

from ....utils.data import DotDict

from .structures import prepare_cache
from .structures import Graph


Node = DotDict  # pylint: disable=invalid-name


class Summarization(object):
    """Summarization algorithm

    Traverses activation tree nodes in preorder.
    Creates graph based on caller_id
    """

    def __init__(self, preorder):
        self.nid = 0
        self.root = None
        self.stack = []
        self.nodes = []
        self.matches = defaultdict(dict)
        self.edges = defaultdict(
            lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(int)))
        )

        self(preorder)

    def graph(self, colors, width=0, height=0):  # pylint: disable=too-many-locals
        """Generate JSON"""
        min_duration = {}
        max_duration = {}
        edges = []
        trials = set()
        for node in self.nodes:
            for trial_id, duration in viewitems(node.duration):
                min_duration[trial_id] = min(
                    min_duration.get(trial_id, float('inf')), duration)
                max_duration[trial_id] = max(
                    max_duration.get(trial_id, float('-inf')), duration)
                trials.add(trial_id)
        for source_nid, targets in viewitems(self.edges):
            for target_nid, types in viewitems(targets):
                for type_, count in viewitems(types):
                    edges.append({
                        'count': count,
                        'source': source_nid,
                        'target': target_nid,
                        'type': type_,
                    })
        tlist = list(trials)
        if not tlist:
            tlist.append(0)
        return {
            'root': self.root,
            'edges': edges,
            'min_duration': min_duration,
            'max_duration': max_duration,
            'colors': colors,
            'trial1': tlist[0],
            'trial2': tlist[-1],
            'width': width,
            'height': height,
        }

    def merge(self, node, activation):
        """Abstract: Merge activation into node"""
        raise NotImplementedError("merge is not implemented")

    def calculate_match(self, node):
        """Abstract: Calculate match. Return tuple"""
        raise NotImplementedError("calculate_match is not implemented")

    def add_edge(self, source, target, type_, count=1):
        """Add edge"""
        ids = target.trial_ids
        trial_id = 0 if len(ids) > 1 else next(iter(ids))
        self.edges[source.index][target.index][type_][trial_id] += count

    def insert_node(self, activation, parent, match=None):
        """Create node for activation

        Arguments:

        activation -- activation element
        parent -- previously created parent node
        match -- matching key
        """
        node = Node(
            index=self.nid,
            parent_index=-1,
            name=activation.name,
            caller_id=activation.caller_id or 0,
            children=[],
            activations=defaultdict(list),
            duration=defaultdict(int),
            full_tooltip=False,
            tooltip=defaultdict(str),
            children_index=-1,
            trial_ids=[],
        )
        self.merge(node, activation)
        self.nid += 1
        if parent is not None:
            node.parent_index = parent.index
            node.children_index = len(parent.children)
            parent.children.append(node)
            if match is not None:
                self.matches[parent.index][match] = node

        self.nodes.append(node)
        return node

    def insert_first(self, call):
        """Insert first node

        Insert node and create initial edge
        """
        self.root = node = self.insert_node(call, None)
        self.add_edge(node, node, 'initial')
        return node

    def insert_call(self, call, last):
        """Insert call

        Insert node, create match, and create call edge
        """
        self.stack.append(last)
        
        match = self.calculate_match(call)
        node = self.matches[last.index].get(match)
        if node is None:
            node = self.insert_node(call, last, match)
        else:
            self.merge(node, call)
        self.add_edge(last, node, 'call')
        return node

    def insert_return(self, last):
        """Insert return

        Create return edge
        """
        temp = self.stack.pop()
        self.add_edge(last, temp, 'return')
        return temp

    def insert_sequence(self, call, last):
        """Insert sequence

        Check if match exists in last.parent. Create node if it doesn't.
        Insert sequence edge from last to call
        """
        match = self.calculate_match(call)
        node = self.matches[last.parent_index].get(match)
        if node is None:
            node = self.insert_node(call, self.nodes[last.parent_index], match)
        else:
            self.merge(node, call)
        self.add_edge(last, node, 'sequence')
        return node

    def __call__(self, preorder):

        for call in preorder:
            if not call.caller_id:
                last = self.insert_first(call)
                continue
            if call.caller_id > last.caller_id:
                last = self.insert_call(call, last)
                continue

            while call.caller_id < last.caller_id:
                last = self.insert_return(last)

            if call.caller_id == last.caller_id:
                last = self.insert_sequence(call, last)

        while self.stack:
            last = self.insert_return(last)

        return self


class LineNameSummarization(Summarization):
    """Summarize Activations by line and name"""
    # ToDo: Diff equivalent

    def merge(self, node, activation):
        """Extract id from activation and insert into idlist"""
        trial_id = activation.trial_id
        if trial_id not in node.trial_ids:
            node.trial_ids.append(trial_id)
        node.activations[trial_id].append(activation.id)
        node.duration[trial_id] += activation.duration

        node.tooltip[trial_id] += "T{} - {}<br>Line {}<br>".format(
            trial_id, activation.id, activation.line
        )

    def calculate_match(self, node):
        """Calculate match. Use line and name"""
        return (node.line, node.name)


class NoMatchSummarization(LineNameSummarization):
    """Create repr for all nodes. Does not summarize tree"""
    # ToDo: Diff equivalent

    def __init__(self, preorder):
        self.match_id = 0
        super(NoMatchSummarization, self).__init__(preorder)

    def calculate_match(self, node):
        """No match"""
        self.match_id += 1
        return self.match_id

    def insert_node(self, activation, parent, match=None):
        """Insert node. Create base repr"""
        node = super(NoMatchSummarization, self).insert_node(
            activation, parent, match
        )
        node.repr = '{0.line}-{0.name}'.format(activation)
        return node

    def insert_call(self, call, last):
        """Insert call.
        Add opening parenthesis to caller"""
        last.repr += "("
        return super(NoMatchSummarization, self).insert_call(call, last)

    def insert_return(self, last):
        """Insert return.
        Add last activation to caller and close parenthesis"""
        parent = super(NoMatchSummarization, self).insert_return(last)
        parent.repr += last.repr + ")"
        return parent

    def insert_sequence(self, call, last):
        """Inser last caller and comma to caller"""
        if not last.children:
            self.nodes[last.parent_index].repr += last.repr + ","
        return super(NoMatchSummarization, self).insert_sequence(call, last)


class StructureSummarization(Summarization):
    """Summarize by substructure"""

    def merge(self, node, activation):
        """Extract ids from activation node and insert into idlist"""
        for trial_id in activation.trial_ids:
            node.activations[trial_id].extend(activation.activations[trial_id])
            node.duration[trial_id] += activation.duration[trial_id]
            node.tooltip[trial_id] += activation.tooltip[trial_id] + "<br>"
            if trial_id not in node.trial_ids:
                node.trial_ids.append(trial_id)

    def calculate_match(self, node):
        """Match by repr"""
        return (node.repr,)

    def __call__(self, preorder):
        return super(StructureSummarization, self).__call__(
            NoMatchSummarization(preorder).nodes
        )


class TreeSummarization(NoMatchSummarization):
    """Build tree"""

    def __call__(self, preorder):
        result = super(TreeSummarization, self).__call__(preorder)
        self.edges.clear()
        stack = [self.root]
        while stack:
            current = stack.pop()
            for index, child in enumerate(current.children):
                self.add_edge(current, child, 'call', index)
                stack.append(child)
        return result


cache = prepare_cache(                                                           # pylint: disable=invalid-name
    lambda self, *args, **kwargs: "trial {}".format(self.trial.id))


class TrialGraph(Graph):
    """Trial Graph Class
       Present trial graph on Jupyter"""

    def __init__(self, trial):
        self.trial = weakref.proxy(trial)

        self.use_cache = True
        self.width = 500
        self.height = 500
        self.mode = 3
        self._modes = {
            0: self.tree,
            1: self.no_match,
            2: self.exact_match,
            3: self.namespace_match
        }

    def result(self, summarization):
        """Get summarization graph result"""
        return self.trial.finished, summarization.graph(
            {self.trial.id: 0}, self.width, self.height
        ), summarization.nodes

    @cache("tree")
    def tree(self):
        """Convert tree structure into dict tree structure"""
        return self.result(TreeSummarization(self.trial.activations))

    @cache("no_match")
    def no_match(self):
        """Convert tree structure into dict graph without node matchings"""
        return self.result(NoMatchSummarization(self.trial.activations))

    @cache("exact_match")
    def exact_match(self):
        """Convert tree structure into dict graph and match equal calls"""
        return self.result(StructureSummarization(self.trial.activations))

    @cache("namespace_match")
    def namespace_match(self):
        """Convert tree structure into dict graph and match namespaces"""
        return self.result(LineNameSummarization(self.trial.activations))

    def _ipython_display_(self):
        from IPython.display import display
        bundle = {
            'application/noworkflow.trial+json': self._modes[self.mode]()[1],
            'text/plain': 'Trial {}'.format(self.trial.id),
        }
        display(bundle, raw=True)



#!/bin/python

import argparse
import sys

def check_color(value):
    if value.upper() in ("YW", "NW"):
        return value.upper()
    if len(value.split(',')) == 3:
        return value
    raise argparse.ArgumentTypeError("%s is an invalid color/schema" % value)


def restyle(data, color="NW", direction="RL", same_line_equal=False):
    new_result = data.replace("rankdir=RL", "rankdir={}".format(direction))
    if same_line_equal:
        new_result = new_result.replace("=\n", "= ")
    call_color = "#3A85B9"
    variable_color = "#85CBD0"
    file_color = "white"
    if color == "YW":
        call_color = "#CCFFCC"
        variable_color = "#FFFFCC"
    elif color != "NW":
        call_color, variable_color, file_color = color.split(",")
    new_result = (
        new_result
        .replace(
            "#3A85B9",
            call_color
        )
        .replace(
            "#85CBD0",
            variable_color
        )
        .replace(
            'fillcolor="white"',
            'fillcolor="{}"'.format(file_color)
        )
    )
    return new_result


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Convert lineage style")
    parser.add_argument("-c", "--color", type=check_color, default="NW",
                        help="Color schema. Possible options: NW, YW or "
                        "three color codes separated by comma, where the "
                        "first one represents calls; the second one "
                        "represents variables; and the third one "
                        "represents files. Ex: '#3A85B9,#FFFFCC,#AAAAAAA'")
    parser.add_argument("-d", "--direction", type=str, default="RL",
                        choices=["BT", "RL", "TB", "LR"],
                        help="Graphviz rankdir. Default=BT")
    parser.add_argument("-e", "--same-line-equal", action='store_true',
                        help="Show values on the same line. \n"
                        "It requires to export the values")

    args = parser.parse_args()
    userinput = "".join(sys.stdin.readlines())
    print(restyle(userinput, color=args.color, direction=args.direction,
          same_line_equal=args.same_line_equal))



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Intermediate Tree Structures and Graph Structures"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)


import json
import time
import traceback

from sqlalchemy import exc

from ... import relational, content
from ...models import GraphCache

from ....utils.cross_version import pickle
from ....utils.io import print_msg


class Graph(object):                                                             # pylint: disable=too-few-public-methods
    """Graph superclass. Handle json transformation"""
    def escape_json(self, data):                                                 # pylint: disable=no-self-use
        """Escape JSON"""
        data = json.dumps(data)
        return (data.replace("&", "\\u0026")
                .replace("<", "\\u003c")
                .replace(">", "\\u003e"))


def prepare_cache(get_type):
    """Decorator: Load graph from cache"""
    def cache(name, attrs=""):
        """Decorator: Load graph from cache"""
        def dec(func):
            """Decorator: Load graph from cache"""
            def load(self, *args, **kwargs):
                """Load graph from cache

                Find graph by type, name and attributes
                If graph is cached, return it

                Return:
                finished -- trial has finished
                graph -- cached trial graph
                """
                cache_session = relational.make_session()

                typ = get_type(self, *args, **kwargs)
                attributes = " ".join(str(kwargs[a])
                                      for a in attrs.split() if a in kwargs)

                information = (typ, name, attributes)
                if self.use_cache:
                    try:
                        caches = GraphCache.select_cache(*information,
                                                         session=cache_session)
                        for cache in caches:
                            result = pickle.loads(
                                content.get(cache.content_hash))
                            if not result[0]:
                                continue
                            cache_session.close()                                # pylint: disable=no-member
                            return result
                    except (ValueError, exc.SQLAlchemyError):
                        traceback.print_exc()
                        print_msg("Couldn't load graph cache", True)
                start = time.time()
                graph = func(self, *args, **kwargs)
                duration = time.time() - start
                try:
                    GraphCache.remove(*information, session=cache_session)
                    GraphCache.create(
                        typ, name, duration, attributes,
                        content.put(pickle.dumps(graph)),
                        session=cache_session, commit=True
                    )
                except exc.SQLAlchemyError:
                    traceback.print_exc()
                    print_msg("Couldn't store graph cache", True)
                cache_session.close()                                            # pylint: disable=no-member
                return graph
            return load
        return dec
    return cache



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Define utility functions"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import inspect
import os

from os.path import join, dirname, exists
from textwrap import dedent
from pkg_resources import resource_string, resource_listdir, resource_isdir


MODULE = __name__
MODULE = MODULE[:MODULE.rfind(".")]
MODULE = MODULE[:MODULE.rfind(".")]
NOWORKFLOW_DIR = dirname(dirname(dirname(__file__)))


def wrap(string, initial="  ", other="\n  "):
    """Re-indent indented text"""
    return initial + other.join(dedent(string).split("\n"))


def resource(filename, encoding=None):
    """Access resource content via setuptools"""
    content = resource_string(MODULE, filename)
    if encoding:
        return content.decode(encoding=encoding)
    return content


def resource_ls(path):
    """Access resource directory via setuptools"""
    return resource_listdir(MODULE, path)


def resource_is_dir(path):
    """Access resource directory via setuptools"""
    return resource_isdir(MODULE, path)


def version():
    """Return noWorkflow version"""
    return resource("../resources/version.txt", encoding="utf-8").strip()


def abstract():
    """Raise abstract Exception"""
    frame = inspect.currentframe().f_back
    name = frame.f_code.co_name
    raise Exception("Abstract method: {}".format(name))


def recursive_copy(origin, destiny):
    """Copy directory from resource to destiny folder"""
    if resource_is_dir(origin):
        if not exists(destiny):
            os.makedirs(destiny)
        for element in resource_ls(origin):
            origin_element = join(origin, element)
            destiny_element = join(destiny, element)
            recursive_copy(origin_element, destiny_element)
    else:
        with open(destiny, "wb") as fil:
            fil.write(resource(origin))



def erase(directory, everything=False):
    """Remove all files from directory


    Keyword Arguments:
    everything -- should delete .noworkflow too (default=False)
    """
    for root, dirs, files in os.walk(directory, topdown=False):
        for name in files:
            path = join(root, name)
            if everything or ".noworkflow" not in path:
                os.remove(join(root, name))
        for name in dirs:
            path = join(root, name)
            if everything or ".noworkflow" not in path:
                os.rmdir(path)



import sys
import numpy as np
from precipitation import read, sum_by_month
from precipitation import write, remove_outliers
from precipitation import create_bargraph

months = np.arange(12) + 1
d13, d14 = read("p13.dat"), read("p14.dat")

for i in range(int(sys.argv[1])):
    write("temp13.dat", remove_outliers(d13), 2013)
    write("temp14.dat", remove_outliers(d14), 2014)
    d13, d14 = read("temp13.dat"), read("temp14.dat")

prec13 = sum_by_month(d13, months)
prec14 = sum_by_month(d14, months)

create_bargraph("out.png", months,
                ["2013", "2014"], prec13, prec14)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""History Graph Module"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import time
import weakref

from collections import OrderedDict, defaultdict

from future.utils import viewvalues

from ....utils.cross_version import zip_longest
from ..trial import Trial
from ..tag import Tag
from .structures import Graph



MAXTRIALS = 1000000
MAX_IN_GRAPH = float("inf")


class HistoryGraph(Graph):
    """History Graph Class

    Present history graph on Jupyter and on command line"""

    cache = {}

    def __init__(self, history, width=500, height=500):
        self.history = weakref.proxy(history)
        self.use_cache = True
        self.width = width
        self.height = height

    def history_data(self):
        """Create history data for presenting graphs

        Return:
        {"nodes": nodes, "edges": edges}
        nodes -- list of trials in filtered history
        edges -- list of edges dicts with keys source and target as node index
        """

        key = (
            self.history.script, self.history.status, self.history.summarize,
            Trial.count()
        )
        if self.use_cache and key in self.cache:
            return self.cache[key]

        tmap = self._load_trials(Trial.reverse_trials(MAXTRIALS))
        graph = self._create_graph(tmap)

        tmap, graph = self._summarize(tmap, graph)

        self._calculate_distances(graph)

        nodes, scripts = self._filter_graph(tmap, graph)

        edges, order, children, actual_graph = self._create_edges(
            graph, nodes, tmap
        )

        self._set_trials_level(tmap, scripts, order, children, actual_graph)

        result = {
            "nodes": nodes,
            "edges": edges,
            "scripts": list(self.history.scripts),
        }
        if self.use_cache:
            self.cache[key] = result

        return result

    def graph(self):
        """Return history_data as a dict graph"""
        result = self.history_data()

        # To JSON
        final = []
        for trial in result["nodes"]:
            dic = trial.to_dict(ignore=("start", "finish"), extra=(
                "level", "status", "tooltip", "duration_text", "code_hash",
                "str_start", "str_finish", "display"
            ))
            final.append(dic)
        return {
            "edges": result["edges"],
            "nodes": final,
            "scripts": result["scripts"],
            "width": self.width,
            "height": self.height,
        }

    def _load_trials(self, trial_gen):  # pylint: disable=no-self-use
        """Preprocess trials


        Add level and tooltip to trials

        Return:
        trials -- preprocessed trials
        ids -- preprocessed trial ids
        tmap -- map trial.id to trial


        Arguments:
        trial_gen -- trial generator
        """
        tmap = OrderedDict()

        for trial in trial_gen:
            trial.display = str(trial.id)
            trial.level = 0
            trial.tooltip = """
                <b>{0.script}</b><br>
                Id: {0.id}<br>
                {status}<br>
                Start: {0.start}<br>
                Finish: {0.finish}
                """.format(trial, status=trial.status.capitalize())
            if trial.finish:
                trial.tooltip += """
                <br>
                Duration: {duration}
                """.format(duration=trial.duration_text)

            tmap[trial.id] = trial

        return tmap

    def _create_graph(self, trial_map):  # pylint: disable=no-self-use
        """Create graph with initial distances

        The graph is represented as a dict of dict of int
        Int values represent the distance between two nodes

        Return:
        graph -- distance graph

        Arguments:
        trial_map -- ordered trial map
        """
        graph = defaultdict(lambda: defaultdict(lambda: MAX_IN_GRAPH))

        for trial in viewvalues(trial_map):
            graph[trial.id][trial.id] = 0
            if trial.parent_id is not None:
                graph[trial.id][trial.parent_id] = 1

        return graph

    def _summarize(self, trial_map, graph):  # pylint: disable=too-many-locals
        """Add display field to trials based on auto tags and summarizes"""
        node_map = OrderedDict()
        new_graph = defaultdict(lambda: defaultdict(lambda: MAX_IN_GRAPH))
        new_tmap = {}

        for tag in Tag.auto_tags():
            if tag.trial_id not in trial_map:
                continue  # Ignore filtered out

            tag_node = Version(tag.name.split('.')[:2])
            trial = trial_map[tag.trial_id]
            trial.display = tag.name
            trial.tooltip = "<b> Trial {}</b><br>{}".format(
                trial.display,
                trial.tooltip
            )
            if tag_node not in node_map:
                node_obj = node_map[tag_node] = Node(tag_node)

                parent_tag = new_tmap.get(trial.parent_id)
                if parent_tag is not None:
                    if (node_obj.parent_id is None or
                            node_obj.parent_id > parent_tag.nid):
                        node_obj.parent_id = parent_tag.nid
            else:
                node_obj = node_map[tag_node]

            new_tmap[tag.trial_id] = node_obj
            node_obj.insert(trial)

        if not self.history.summarize:
            return trial_map, graph

        node_map = OrderedDict(reversed(list(node_map.items())))

        new_graph = defaultdict(lambda: defaultdict(lambda: MAX_IN_GRAPH))

        for node in viewvalues(node_map):
            new_graph[node.id][node.id] = 0

        for origin, distances in graph.items():
            if origin not in new_tmap:
                continue
            orignode = new_tmap[origin].id
            for target, dist in distances.items():
                if target not in new_tmap:
                    continue
                targnode = new_tmap[target].id
                new_graph[orignode][targnode] = min(
                    new_graph[orignode][targnode], dist
                )

        return node_map, new_graph

    def _calculate_distances(self, graph):  # pylint: disable=no-self-use
        """Prepare history graph


        The graph is represented as a dict of dict of int
        Int values represent the distance between two nodes
        Use Floyd-Warshall algorithm to calculate distances


        The returned graph can be used to:
        -find parent trial after filter
        -find previous trials in a branch line


        Arguments:
        graph -- distance graph
        """
        for k in graph:
            for i in graph:
                for j in graph:
                    if graph[i][j] > graph[i][k] + graph[k][j]:
                        graph[i][j] = graph[i][k] + graph[k][j]

    def _filter_graph(self, trial_map, graph):
        """Filter history graph

        Applies script and status filters on the graph
        Filters remove trials that do not match the conditions from the graph


        Return:
        nodes -- filtered trials
        scripts -- group trials by scripts


        Arguments:
        trial_map -- ordered trial map
        graph -- distance graph
        """
        status = self.history.status.lower()
        script = self.history.script
        nodes = []
        scripts = defaultdict(list)
        nid = 0
        for trial in reversed(list(trial_map.values())):
            if not trial.match_status(status) or not trial.match_script(script):
                for tid in trial_map:
                    graph[tid][trial.id] = MAX_IN_GRAPH
            else:
                nodes.append(trial)
                trial.nid = nid
                scripts[trial.script].append(trial)
                nid += 1

        return nodes, scripts

    def _create_edges(self, graph, nodes, trial_map):
        """Create edges for graph

        Arguments:
        graph -- dict of dict of int with min distances between trials
        nodes -- list of nodes from the oldest to the newest
        trial_map -- map of trial.id to trial node


        Return:
        edges -- edge list of dicts with source and target keys
        order -- ordered dict with desired script order
        children -- dict with list of trials that have key trial as target
        actual_graph -- edge dict by trial id
        """

        edges = []
        order = OrderedDict()
        children = defaultdict(list)
        actual_graph = {}

        for source, target in self._edges(graph, nodes, script_order=order):
            edges.append({
                "source": trial_map[source].nid,
                "target": trial_map[target].nid,
                "right": 1,
                "level": 0
            })
            actual_graph[source] = target
            children[target].append(source)

        return (edges, order, children, actual_graph)

    def _edges(self, graph, nodes, script_order=None):  # pylint: disable=no-self-use
        """Edge generator. Iterate through all edges on graph


        Arguments:
        graph -- dict of dict of int with min distances between trials
        nodes -- list of nodes from the oldest to the newest

        Keyword arguments:
        script_order -- ordered dict to be touched for setting the script order
        """
        if script_order is None:
            script_order = {}

        for trial in reversed(nodes):
            tid = trial.id
            target = min(
                graph[tid],
                key=lambda x, i=tid: float("inf") if x == i else graph[i][x]
            )
            if graph[tid][target] != MAX_IN_GRAPH:
                yield (tid, target)
            script_order[trial.script] = 1

    def _set_trials_level(self, tmap, scripts, order, children, actual_graph):  # pylint: disable=no-self-use, too-many-arguments
        """Adjust levels of trials according to their script and branchs


        Arguments:
        tmap -- map of trial.id to trial
        scripts -- group trials by scripts
        order -- ordered dict with desired script order
        children -- dict with list of trials that have key trial as target
        actual_graph -- edge dict by trial id
        """

        previous_lines = [[0, None]]

        def add_script(script, level, min_id, max_id):
            """Find an appropriate line to include script
            Naive implementation: considers only the boundaries of previous lines"""
            if min_id == max_id and previous_lines[0][0] == 0:
                lines = iter(previous_lines)
                line = next(lines)
                line[0] = 1
                for line in lines:
                    line[0] += 1
                    for trial in scripts[line[1]]:
                        trial.level += 1
            elif min_id != max_id:
                total = previous_lines[-1][0]
                previous_lines.append(
                    [total + level, script]
                )
                for trial in scripts[script]:
                    trial.level += total

        for script in order:
            level = 0
            min_id = Version.as_version(MAX_IN_GRAPH + 1)
            max_id = Version.as_version(-MAX_IN_GRAPH - 1)
            for trial in scripts[script]:
                min_id = min(min_id, Version.as_version(trial.id))
                max_id = max(max_id, Version.as_version(trial.id))
                if trial.id not in actual_graph:
                    # trial is isolated
                    trial.level = level
                    level += 1
                    continue

                parent_id = actual_graph[trial.id]
                if children[parent_id].index(trial.id) > 0:
                    # trial is not the first child
                    # increase level
                    trial.level = level
                else:
                    # trial is the first child
                    # use the parent level
                    parent = tmap[parent_id]
                    if parent.id == trial.id:
                        trial.level = level
                    else:
                        trial.level = parent.level
                level = max(level, trial.level + 1)
            add_script(script, level, min_id, max_id)

    def _ipython_display_(self):
        from IPython.display import display
        bundle = {
            'application/noworkflow.history+json': self.graph(),
            'text/plain': repr(self),
        }
        display(bundle, raw=True)

    def __repr__(self):
        width = max(len(s) for s in self.history.scripts)
        history = self.history_data()
        nodes = history["nodes"]

        if not nodes:
            return ""

        to_level = {
            nodes[edge["source"]].id: nodes[edge["target"]].level
            for edge in history["edges"]
        }

        max_level = max(t.level for t in nodes)
        active_levels = [0] * (max_level + 1)
        lines = []
        add_line = lambda *a, **kw: lines.append(_line_text(*a, **kw))

        for trial in reversed(nodes):
            active_levels[trial.level] = 1
            add_line(active_levels, trial, trial.level, width=width)

            if trial.id not in to_level:
                # First trial for script
                active_levels[trial.level] = 0

            elif to_level[trial.id] != trial.level:
                # Start of new branch
                active_levels[trial.level] = 0
                for i in range(trial.level, to_level[trial.id], -1):
                    add_line(active_levels, trial, i, moving=True, width=width)
            if trial.parent_id is None:
                active_levels[trial.level] = 0
            lines.append(_blank_text(active_levels))


        return "\n".join(lines)


class Node(object):
    """Node object with specific fields for graph"""

    def __init__(self, tid):
        self.id = tid
        self.nid = None
        self.parent_id = None
        self.level = 0
        self.script = ""
        self.trials = []

    def insert(self, trial):
        """Insert trial to Node"""
        self.trials.append(trial)
        self.script = trial.script

    def match_status(self, status):
        """Match statuses in node"""
        new_trials = [
            trial for trial in self.trials
            if trial.match_status(status)
        ]
        if new_trials:
            self.trials = new_trials
            return True
        return False

    def match_script(self, script):
        """Match scripts in node"""
        new_trials = [
            trial for trial in self.trials
            if trial.match_script(script)
        ]
        if new_trials:
            self.trials = new_trials
            return True
        return False

    def to_dict(self, *args, **kwargs):
        """Convert to dict"""
        return {
            'id': repr(self.id),
            "display": repr(self.id),
            'parent_id': self.parent_id,
            'level': self.level,
            'trials': [x.to_dict(*args, **kwargs) for x in self.trials],
        }


class Version(object):
    """Represents a version number"""

    def __init__(self, numbers):
        self.numbers = tuple(numbers)

    def __gt__(self, other):  # pylint: disable=too-many-return-statements
        for first, second in zip_longest(self.numbers, other.numbers):
            if first == second:
                continue
            if first is None:
                return False
            if second is None:
                return True
            if isinstance(first, str) and first.startswith('b'):
                if isinstance(second, str) and second.startswith('b'):
                    return float(first[1:]) > float(second[1:])
                return False
            if isinstance(second, str) and second.startswith('b'):
                return True
            return float(first) > float(second)

        return False

    def __hash__(self):
        return hash(self.numbers)

    def __eq__(self, other):
        return self.numbers == other.numbers

    def __repr__(self):
        return '.'.join(map(str, self.numbers))

    @classmethod
    def as_version(cls, number):
        """Convert version number to version object"""
        if isinstance(number, cls):
            return number
        return cls([number])


def _line_text(active, trial, current, moving=False, width=25):
    """Return text for line history"""
    text = []
    for i, value in enumerate(active):
        if i == current:
            text.append("/ " if moving else
                        " {}".format(trial.status_letter))
        elif value:
            text.append(" |")
        else:
            text.append("  ")
    if moving:
        return "".join(text)
    return "{line}  {id: <4} {script: <{width}} {tags}".format(
        line="".join(text), id=trial.id, script=trial.script,
        tags=", ".join(tag.name for tag in trial.tags), width=width
    )


def _blank_text(active):
    """Return text for line history"""
    text = []
    for value in active:
        if value:
            text.append(" |")
        else:
            text.append("  ")
    return "".join(text)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""'now helper' command"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os
import subprocess
import sys

from ..utils.functions import resource, resource_ls
from ..utils.functions import recursive_copy, erase


from .command import Command


HELPER = "../resources/helper"


class Helper(Command):
    """Create a Helper for noWorkflow usage"""

    def add_arguments(self):
        add_arg = self.add_argument
        choices = sorted(list(resource_ls(HELPER)))
        add_arg("id", type=str, nargs="?", choices=choices,
                help="helper identification")
        add_arg("--dir", type=str,
                help="set helper path. Default to CWD/<id>"
                     "where <id> is the helper identification")

    def execute(self, args):
        directory = "{}".format(args.id)
        if args.dir:
            directory = args.dir
        print("Creating Helper {}".format(args.id))
        helper_path = os.path.join(HELPER, args.id)
        recursive_copy(helper_path, directory)



#!/usr/bin/python2

import csv
import numpy as np
import matplotlib.pyplot as plt
import time

from itertools import chain
from collections import defaultdict


def read(filename):
    result = defaultdict(list)
    with open(filename, "r") as c:
        reader = csv.reader(c, delimiter=";")
        for row in reader:
            month = int(row[1].split("/")[1])
            precipitation = float(row[3])
            result[month].append(precipitation)
    return result

def write(filename, data, year):
    with open(filename, "w") as c:
        writer = csv.writer(c, delimiter=";")
        for month in sorted(data.keys()):
            for day, value in enumerate(data[month]):
                writer.writerow([
                    83743, "{:02}/{:02}/{}".format(day + 1, month, year),
                    1200, value])

def remove_outliers(data, thresh=2.5):
    full_data = np.asarray(tuple(chain.from_iterable(data[i]
                                 for i in sorted(data.keys()))))
    non_zeros = full_data != 0
    median = np.median(full_data[non_zeros])
    result = {}
    for month in data:
        values = np.asarray(data[month])[:, None]
        diff = np.sum((values - median)**2, axis=-1)
        diff = np.sqrt(diff)
        med_abs_deviation = np.median(diff)

        modified_z_score = 0.6745 * diff / med_abs_deviation

        outliers = modified_z_score > thresh
        non_outliers = modified_z_score <= thresh

        new_data = np.zeros(len(values))
        new_data[non_outliers] = np.transpose(values[non_outliers])[0]
        new_data[outliers] = median

        result[month] = new_data.tolist()
    return result


def prepare(series, months, names, div=.1, colors=["b", "g", "r"]):
    fig, ax = plt.subplots()

    ax.set_ylabel("Precipitation (mm)")
    ax.set_xlabel("Month")
    ax.set_title("Precipitation by Month")
    ax.set_xticks(months + .5)
    ax.set_xticklabels(list(map(str, months)))
    ax.set_ylim([0, 400])

    half_div = div / 2.0
    width = (1.0 - div) / len(series)
    bars = []
    for i, data in enumerate(series):
        bars.append(ax.bar(months + half_div + i * width, data, width,
                           color=colors[i]))

    ax.legend(bars, names)

def create_bargraph(output, months, years, *prec):
    prepare(prec, months, years)
    plt.savefig(output)

def sum_by_month(data, months):
    time.sleep(1)
    return [sum(data[i]) for i in months]

__VERSION__ = "1.1.0"

if __name__ == "__main__":
    import sys
    filename = sys.argv[1]
    month = sys.argv[2]
    data = read(filename)
    print(";".join(map(str, data[int(month)])))



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Define views for 'now vis'"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os

from flask import render_template, jsonify, request

from ..persistence.models import Trial
from ..persistence.models.history import History
from ..persistence.models.diff import Diff
from ..persistence import relational


class WebServer(object):
    """Flask WebServer"""
    # pylint: disable=too-few-public-methods
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(WebServer, cls).__new__(
                cls, *args, **kwargs)
        return cls._instance

    def __init__(self):
        from flask import Flask

        self.app = Flask(__name__)


app = WebServer().app  # pylint: disable=invalid-name


@app.after_request
def add_header(req):
    """
    Add headers to both force latest IE rendering engine or Chrome Frame,
    and also to cache the rendered page for 10 minutes.
    """
    req.headers["Cache-Control"] = "no-cache, no-store, must-revalidate"
    req.headers["Pragma"] = "no-cache"
    req.headers["Expires"] = "0"
    req.headers['Cache-Control'] = 'public, max-age=0'
    return req


@app.route("/<path:path>")
def static_proxy(path):
    """Serve static files"""
    return app.send_static_file(path)


@app.route("/")
@app.route("/<tid>-<graph_mode>")  # todo
def index(tid=None, graph_mode=None):
    """Respond history scripts and index page as HTML"""
    # pylint: disable=unused-argument
    history = History()
    return render_template(
        "index.html",
        cwd=os.getcwd(),
        scripts=history.scripts
    )


@app.route("/trials.json")
@app.route("/trials") # remove
def trials():
    """Respond history graph as JSON"""
    history = History(script=request.args.get("script"),
                      status=request.args.get("execution"),
                      summarize=bool(int(request.args.get("summarize"))))
    return jsonify(**history.graph.graph())


@app.route("/trials/<tid>/<graph_mode>/<cache>.json")
def trial_graph(tid, graph_mode, cache):
    """Respond trial graph as JSON"""
    trial = Trial(tid)
    graph = trial.graph
    graph.use_cache &= bool(int(cache))
    _, tgraph, _ = getattr(graph, graph_mode)()
    return jsonify(**tgraph)


@app.route("/trials/<tid>/dependencies.json")
@app.route("/trials/<tid>/dependencies")  # remove
def dependencies(tid):
    """Respond trial module dependencies as JSON"""
    # pylint: disable=not-an-iterable
    trial = Trial(tid)
    local = [x.to_dict(extra=("code_hash",)) for x in trial.local_modules]
    result = [x.to_dict(extra=("code_hash",)) for x in trial.modules]
    return jsonify(local=local, all=result)


@app.route("/trials/<tid>/environment.json")
@app.route("/trials/<tid>/environment")  # remove
def environment(tid):
    """Respond trial environment variables as JSON"""
    trial = Trial(tid)
    return jsonify(env=trial.environment)


@app.route("/trials/<tid>/file_accesses.json")
@app.route("/trials/<tid>/file_accesses")  # remove
def file_accesses(tid):
    """Respond trial file accesses as JSON"""
    trial = Trial(tid)
    return jsonify(file_accesses=[x.to_dict(extra=("stack",))
                                  for x in trial.file_accesses])


@app.route("/diff/<trial1>/<trial2>/info.json")
def diff(trial1, trial2):
    """Respond trial diff as JSON"""
    diff_object = Diff(trial1, trial2)
    return jsonify(
        trial1=diff_object.trial1.to_dict(extra=("duration_text",)),
        trial2=diff_object.trial2.to_dict(extra=("duration_text",)),
        trial=diff_object.trial,
    )

@app.route("/diff/<trial1>/<trial2>/dependencies.json")
def diff_modules(trial1, trial2):
    """Respond modules diff as JSON"""
    diff_object = Diff(trial1, trial2)
    modules_added, modules_removed, modules_replaced = diff_object.modules
    return jsonify(
        modules_added=[x.to_dict(extra=("code_hash",)) for x in modules_added],
        modules_removed=[x.to_dict(extra=("code_hash",)) for x in modules_removed],
        modules_replaced=[[y.to_dict(extra=("code_hash",)) for y in x] for x in modules_replaced],
    )

@app.route("/diff/<trial1>/<trial2>/environment.json")
def diff_environment(trial1, trial2):
    """Respond environment diff as JSON"""
    diff_object = Diff(trial1, trial2)
    env_added, env_removed, env_replaced = diff_object.environment
    return jsonify(
        env_added=[x.to_dict() for x in env_added],
        env_removed=[x.to_dict() for x in env_removed],
        env_replaced=[[y.to_dict() for y in x] for x in env_replaced],
    )

@app.route("/diff/<trial1>/<trial2>/file_accesses.json")
def diff_accesses(trial1, trial2):
    """Respond trial diff as JSON"""
    diff_object = Diff(trial1, trial2)
    fa_added, fa_removed, fa_replaced = diff_object.file_accesses
    return jsonify(
        fa_added=[x.to_dict() for x in fa_added],
        fa_removed=[x.to_dict() for x in fa_removed],
        fa_replaced=[[y.to_dict() for y in x] for x in fa_replaced],
    )


@app.route("/diff/<trial1>/<trial2>/<graph_mode>-<cache>.json")
def diff_graph(trial1, trial2, graph_mode, cache):
    """Respond trial diff as JSON"""
    diff_object = Diff(trial1, trial2)
    graph = diff_object.graph
    graph.use_cache &= bool(int(cache))

    _, diff_result, _ = getattr(graph, graph_mode)()
    return jsonify(**diff_result)


@app.teardown_appcontext
def shutdown_session(exception=None):
    """Shutdown SQLAlchemy session"""
    # pylint: disable=unused-argument
    relational.session.remove()



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""'%%now_sql' magic"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)


from future.utils import viewkeys, viewvalues, text_to_native_str
from IPython.core.display import display_javascript
from IPython.utils.text import DollarFormatter

from ...persistence import relational
from ...utils.formatter import Table

from .command import IpythonCommandMagic


class NowSQL(IpythonCommandMagic):
    """Query the provenance database with SQL

    Examples
    --------
    ::
        In  [1]: %%now_sql
            ...: SELECT DISTINCT script FROM trial
        Out [1]: [['script'],
            ...:  ['script1.py'],
            ...:  ['script2.py']]

        In  [2]: %%now_sql --result tupleit
            ...: SELECT * FROM trial

        In  [3]: tupleit
        Out [3]: <generator object query at 0x7fd9ebdbebe0>
    """

    def __init__(self, *args, **kwargs):
        super(NowSQL, self).__init__(*args, **kwargs)
        javascript = """
            var mode = 'magic_text/x-sql';
            if (!Jupyter.CodeCell.options_default.highlight_modes[mode]) {
                Jupyter.CodeCell.options_default.highlight_modes[mode] = {
                    'reg':[]
                };
            }
            Jupyter.CodeCell.options_default.highlight_modes[mode].reg.push(
                /^%%now_sql/
            );
        """
        display_javascript(javascript, raw=True)

    def add_arguments(self):
        super(NowSQL, self).add_arguments()
        add_arg = self.add_argument
        add_arg("--result", type=str,
                help="""The variable in which the result will be stored""")

    def execute(self, func, line, cell, magic_cls):
        formatter = DollarFormatter()
        cell = formatter.vformat(cell, args=[],
                                 kwargs=magic_cls.shell.user_ns.copy())
        _, args = self.arguments(func, line)
        result = relational.query(text_to_native_str(cell))
        if args.result:
            magic_cls.shell.user_ns[args.result] = result
        else:
            result = list(result)
            table = Table()
            if result:
                table.append(list(viewkeys(result[0])))
            for line in result:
                table.append(list(viewvalues(line)))
            return table



# Copyright (c) 2017 Universidade Federal Fluminense (UFF)
# Copyright (c) 2017 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Diff Graph Module"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import weakref

from copy import copy
from collections import defaultdict
from functools import cmp_to_key

from apted import meta_chained_config, Config, APTED
from future.utils import viewitems

from .trial_graph import Node
from .structures import Graph, prepare_cache


class NowConfig(Config):
    """APTED configuration"""

    def rename(self, node1, node2):
        """Calculates the cost of renaming the label of the source node
        to the label of the destination node"""
        if node1.parent_index == node2.parent_index == -1:
            return 0
        return 3 if node1.name != node2.name else 0


CONFIG = meta_chained_config(NowConfig)()


@cmp_to_key
def children_cmp(node1, node2):
    """Compares children node for sorting"""
    if node1.original1 is not None and node2.original1 is not None:
        return node1.original1 - node2.original1
    elif node1.original2 is not None and node2.original2 is not None:
        return node1.original2 - node2.original2
    return -1


def merge(node1, node2, id_to_node1, id_to_node2):
    """Creates a merged node based on node1 combined to node2"""
    new_node = copy(node1)
    new_node.children1 = node1.children
    new_node.children2 = node2.children
    new_node.activations.update(node2.activations)
    new_node.duration.update(node2.duration)
    new_node.tooltip.update(node2.tooltip)
    new_node.trial_ids.extend(node2.trial_ids)
    new_node.full_tooltip &= node2.full_tooltip
    new_node.original1 = node1.index
    new_node.original2 = node2.index
    id_to_node1[node1.index] = new_node
    id_to_node2[node2.index] = new_node
    return new_node


def create_mapping(root1, root2):
    """Creates mapping between trees rooted at root1 and root2

    Returns:
    -- new root
    -- map from node index 1 to resulting node
    -- map from node index 2 to resulting node
    """
    apted = APTED(root1, root2, CONFIG)
    mapping = apted.compute_edit_mapping()

    combined_duration = copy(root1.duration)
    combined_duration.update(root2.duration)
    trial_ids = root1.trial_ids + root2.trial_ids
    id_to_node1 = {}
    id_to_node2 = {}

    for node1, node2 in mapping:
        if node1 is None:
            node = id_to_node2[node2.index] = copy(node2)
            node.children1 = []
            node.children2 = node2.children
            node.original1 = None
            node.original2 = node2.index

        elif node2 is None:
            node = id_to_node1[node1.index] = copy(node1)
            node.children1 = node1.children
            node.children2 = []
            node.original1 = node1.index
            node.original2 = None
        else:
            if node1.name != node2.name:
                print("Warning. Mismatch?")
            merge(node1, node2, id_to_node1, id_to_node2)
            # Note that it overrides node1 attributes

    if id_to_node1[root1.index] is not id_to_node2[root2.index]:
        root = Node(
            index=0,
            parent_index=0,
            name="<diff>",
            caller_id=0,
            original1=None,
            original2=None,
            children1=[root1],
            children2=[root2],
            activations=[],
            duration=combined_duration,
            full_tooltip=True,
            tooltip={x: "Diff" for x in trial_ids},
            children_index=-1,
            trial_ids=trial_ids
        )
    else:
        root = id_to_node1[root1.index]

    return root, id_to_node1, id_to_node2


def merge_edges(edges1, edges2, id_to_node1, id_to_node2):
    """Merge edges"""
    dic = defaultdict(
        lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(int)))
    )
    for edge in edges1:
        source = id_to_node1[edge['source']].index
        target = id_to_node1[edge['target']].index
        dic[source][target][edge['type']].update(edge['count'])
    for edge in edges2:
        source = id_to_node2[edge['source']].index
        target = id_to_node2[edge['target']].index
        dic[source][target][edge['type']].update(edge['count'])
    edges = []
    for source_nid, targets in viewitems(dic):
        for target_nid, types in viewitems(targets):
            for type_, count in viewitems(types):
                edges.append({
                    'count': count,
                    'source': source_nid,
                    'target': target_nid,
                    'type': type_,
                })
    return edges


def create_diff(trial_graph1, trial_graph2):
    """Creates a graph structure that combines both graphs"""
    # pylint: disable=too-many-locals
    finished1, graph1, _ = trial_graph1
    finished2, graph2, _ = trial_graph2
    root, id_to_node1, id_to_node2 = create_mapping(
        graph1['root'], graph2['root']
    )

    nodes = []
    index = 1
    queue = [root]
    visited = {id(root)}
    visited_queue = set()
    while queue:
        node = queue.pop()
        nodes.append(node)
        node.index = index
        node.children = []
        for ochild in node.children1:
            child = id_to_node1[ochild.index]
            if id(child) not in visited:
                node.children.append(child)
                child.caller_id = child.parent_index = index
                visited.add(id(child))
        for ochild in node.children2:
            child = id_to_node2[ochild.index]
            if id(child) not in visited:
                node.children.append(child)
                child.caller_id = child.parent_index = index
                visited.add(id(child))
        del node.children1
        del node.children2
        node.children.sort(key=children_cmp)

        lchildren = len(node.children) - 1
        for cindex, child in enumerate(reversed(node.children)):
            if id(child) not in visited_queue:
                queue.append(child)
            child.children_index = lchildren - cindex
            visited_queue.add(id(child))
        index += 1

    graph = {
        'root': root,
        'edges': merge_edges(
            graph1['edges'], graph2['edges'],
            id_to_node1, id_to_node2
        ),
        'min_duration': {},
        'max_duration': {},
        'colors': {
            graph1["trial1"]: 1,
            graph2["trial1"]: 2,
        },
        'trial1': graph1["trial1"],
        'trial2': graph2["trial1"],
        'width': graph1['width'],
        'height': graph1['height'],
    }
    graph['min_duration'].update(graph1['min_duration'])
    graph['min_duration'].update(graph2['min_duration'])
    graph['max_duration'].update(graph1['max_duration'])
    graph['max_duration'].update(graph2['max_duration'])
    finished = finished1 and finished2
    return finished, graph, nodes


cache = prepare_cache(  # pylint: disable=invalid-name
    lambda self, *args, **kwargs: "diff {}:{}".format(self.diff.trial1.id,
                                                      self.diff.trial2.id))

class DiffGraph(Graph):
    """Diff Graph Class. Present diff graph on Jupyter"""

    def __init__(self, diff):
        self.diff = weakref.proxy(diff)
        self.use_cache = False
        self.width = 500
        self.height = 500

        self.mode = 2

        self._modes = {
            0: self.tree,
            1: self.no_match,
            2: self.exact_match,
            3: self.namespace_match
        }

    @cache("tree")
    def tree(self):
        """Convert tree structure into dict tree structure"""
        return create_diff(
            self.diff.trial1.graph.tree(),
            self.diff.trial2.graph.tree()
        )

    @cache("no_match")
    def no_match(self):
        """Convert tree structure into dict graph without node matchings"""
        return create_diff(
            self.diff.trial1.graph.no_match(),
            self.diff.trial2.graph.no_match()
        )

    @cache("exact_match")
    def exact_match(self):
        """Convert tree structure into dict graph and match equal calls"""
        return create_diff(
            self.diff.trial1.graph.exact_match(),
            self.diff.trial2.graph.exact_match()
        )

    @cache("namespace_match")
    def namespace_match(self):
        """Convert tree structure into dict graph and match namespaces"""
        return create_diff(
            self.diff.trial1.graph.namespace_match(),
            self.diff.trial2.graph.namespace_match()
        )

    def _ipython_display_(self):
        from IPython.display import display
        bundle = {
            'application/noworkflow.trial+json': self._modes[self.mode]()[1],
            'text/plain': 'Diff {}:{}'.format(
                self.diff.trial1.id,
                self.diff.trial2.id
            ),
        }
        display(bundle, raw=True)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""'now demo' command"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os
import subprocess
import sys

from ..utils.functions import resource, resource_ls
from ..utils.functions import recursive_copy, erase


from .command import Command


DEMO = "../resources/demo"


class Demo(Command):
    """Create demo project"""

    def add_arguments(self):
        add_arg = self.add_argument
        choices = sorted(list(resource_ls(DEMO)))
        add_arg("id", type=str, nargs="?", choices=choices,
                help="demo identification")
        add_arg("--steps", type=str, default=":",
                help="select step slice to run. Format start:stop:step")
        add_arg("--dir", type=str,
                help="set demo path. Default to CWD/demo<id>"
                     "where <id> is the demo identification")

    def execute(self, args):
        directory = "demo_{}".format(args.id)
        if args.dir:
            directory = args.dir
        print("Creating Demo {}".format(args.id))
        demo_path = os.path.join(DEMO, args.id)
        steps = resource(os.path.join(demo_path, "steps.txt"),
                         encoding="utf-8")
        steps = steps.split("\n")
        steps = steps[slice(*[int(x) if x != "" else None
                              for x in args.steps.split(":")])]
        for step in steps:
            print(step)
            words = step.split(" ")
            param = " ".join(words[1:])
            if words[0] == ">LOAD":
                recursive_copy(os.path.join(demo_path, param),
                               directory)
            if words[0] == ">ERASE":
                erase(directory)
            if words[0] == ">ERASE_ALL":
                erase(directory, everything=True)
            if words[0] == "$now":
                cwd = os.getcwd()
                os.chdir(directory)
                subprocess.call("{} {}".format(sys.argv[0], param),
                                shell=True, stdout=subprocess.PIPE)
                os.chdir(cwd)
        os.chdir(directory)
